// -*- C++ -*-

/* 
 * plotlayout.cc
 *
 * Copyright (C) 1998 EMC Capital Management, Inc.
 *
 * Developed by Jon Trowbridge <trow@emccta.com> and
 * Havoc Pennington <hp@emccta.com>.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "plotlayout.h"

#include "grid.h"

#include <iostream>

/////////////////////////////////////////////////////

class PlotLayoutImpl : public Gtk_Widget {
public:
  PlotLayoutImpl();
  ~PlotLayoutImpl(); 

  // blow everything away.
  void clear(); 

  // These calls destroy any existing plot or axis
  void place_plot(Gtk_Widget* w, size_t row, size_t col);
  void place_axis(Gtk_Axis* w, size_t row, size_t col, Axis_Base::alignment_t a);

  // Destroy everything on a row or column and shift
  //  other rows/columns accordingly.
  void remove_row(size_t row);
  void remove_column(size_t col);
  
  // These don't shift other things, they just leave a blank.
  void erase_axis(size_t row, size_t col, Axis_Base::alignment_t a);
  void erase_plot(size_t row, size_t col);
  void erase_axes(size_t row, size_t col);
  // delete both axes and plots
  void erase_group(size_t row, size_t col);  
  
  void move_axis (size_t row1, size_t col1, size_t row2, size_t col2, 
		  Axis_Base::alignment_t a);
  void move_plot (size_t row1, size_t col1, size_t row2, size_t col2);
  void move_group(size_t row1, size_t col1, size_t row2, size_t col2);
  void swap_group(size_t row1, size_t col1, size_t row2, size_t col2);

  // Check on state
  bool plot_exists(size_t row, size_t col);
  bool axis_exists(size_t row, size_t col, Axis_Base::alignment_t a);
  bool group_exists(size_t row, size_t col);

  // per-plot title
  void set_title(size_t row, size_t col, const string & title);
  // whole-grid title.
  void set_title(const string & title);

  Gtk_Widget* get_plot(size_t row, size_t col);
  Gtk_Axis*   get_axis(size_t row, size_t col, Axis_Base::alignment_t a);

  size_t rows() const; 
  size_t columns() const; 

protected:
  virtual void paint(GdkRectangle* area);
  virtual void size_request_impl(GtkRequisition* r);
  virtual void size_allocate_impl(GtkAllocation* a);
  virtual void realize_impl();
  virtual void map_impl();
  virtual void unmap_impl();
  virtual void draw_impl(GdkRectangle* area);
  virtual gint expose_event_impl(GdkEventExpose* e);
  virtual gint configure_event_impl(GdkEventConfigure* e);

  void setup_child(Gtk_Widget* child);

private:
  class Group {
  public:
    Group() : plot_(0) { 
      size_t i = 0; 
      while ( i < 4 ) { axes_[i] = 0; ++i;} 
    }
    ~Group() { clear(); }
    
    void clear() { 
      set_plot(0);
      erase_axes();
    }

    Gtk_Axis* get_axis(Axis_Base::alignment_t a) { return axes_[a]; }
    void      replace_axis(Gtk_Axis* axis, Axis_Base::alignment_t a) {
      delete axes_[a];
      axes_[a] = axis;
    }

    void      set_axis(Gtk_Axis* axis, Axis_Base::alignment_t a) {
      axes_[a] = axis;
    }

    Gtk_Widget* get_plot() { return plot_; }
    void        replace_plot(Gtk_Widget* w) {
      delete plot_;
      plot_ = w;
    }

    void        set_plot(Gtk_Widget* w) {
      plot_ = w;
    }
      
    void erase_axes() {
      size_t i = 0; 
      while ( i < 4 ) {
	delete axes_[i];
	axes_[i] = 0;
	++i;
      }
    }

#ifdef GNOME_ENABLE_DEBUG
    void spew(size_t r, size_t c) {
      g_print("[%2u,%2u p: %p n: %p s: %p e: %p w: %p]",
	      r,c,plot_,axes_[Axis_Base::NORTH],axes_[Axis_Base::SOUTH],
	      axes_[Axis_Base::EAST],axes_[Axis_Base::WEST]);
    }
#endif

  private:
    Gtk_Widget* plot_;
    // axes indexed by alignment enum.
    Gtk_Axis* axes_[4];
  };

  Grid<Group*> groups_;

  bool in_range(size_t row, size_t col) {
    if (row >= rows() || col >= columns()) {
      return false;
    }    
    else return true;
  }

  void ensure_exists(size_t row, size_t col) {
    // make sure we have space
    if (!in_range(row,col)) {
      size_t new_rows = MAX(rows(),row+1); 
      size_t new_cols = MAX(columns(),col+1); 
#ifdef GNOME_ENABLE_DEBUG
      g_print("Reserving rows: %u cols: %u\n", new_rows, new_cols);
#endif
      groups_.reserve(new_rows,new_cols);
    }    

    // make sure we have a group.
    if (groups_[row][col] == 0) {
      groups_[row][col] = new Group;
    }
  }

#ifdef GNOME_ENABLE_DEBUG
    void spew() {
      g_print("PlotLayout contents:\n");
      size_t r = rows();
      size_t c = columns();
      size_t i = r;
      while ( true ) {
	if (i == 0) break;
	--i;

	size_t j = 0;
	while ( j < c ) {
	  if ( groups_[i][j] != 0 ) {
	    groups_[i][j]->spew(i,j);
	  }
	  else g_print("[       0        ]");

	  g_print(" ");
	  ++j;
	}
	g_print("\n");
      }
    }      
#endif
};


PlotLayout::PlotLayout()
  : Gtk_Frame(),
    impl_(new PlotLayoutImpl)
{
  Gtk_Container::add(impl_);
  impl_->show();
}

PlotLayout::~PlotLayout()
{
  delete impl_;
}


PlotLayoutImpl::PlotLayoutImpl()
{
  // Eventually we probably want this to be configurable.
  GtkStyle* style = gtk_style_copy(get_style()->gtkstyle);
  GdkColor color;
  gdk_color_parse("white",&color);
  style->bg[GTK_STATE_NORMAL] = color;
  set_style(style);
}

PlotLayoutImpl::~PlotLayoutImpl()
{
  clear();
}

static inline GtkRequisition & REQUISITION(Gtk_Widget* w) 
{
#ifdef GNOME_ENABLE_DEBUG
  static GtkRequisition r; r.width = 0; r.height = 0;
  g_return_val_if_fail(w != 0, r);
  g_return_val_if_fail(w->gtkobj() != 0, r);
  g_return_val_if_fail(GTK_IS_WIDGET(w->gtkobj()), r);
#endif
  return (GTK_WIDGET(w->gtkobj())->requisition);
}

static inline GtkAllocation & ALLOCATION(Gtk_Widget* w) 
{
#ifdef GNOME_ENABLE_DEBUG
  static GtkAllocation a; a.width = 0; a.height = 0; a.x = 0; a.y = 0;
  g_return_val_if_fail(w != 0, a);
  g_return_val_if_fail(w->gtkobj() != 0, a);
  g_return_val_if_fail(GTK_IS_WIDGET(w->gtkobj()), a);
#endif
  return (GTK_WIDGET(w->gtkobj())->allocation);
}

static bool operator==(const GtkAllocation & lhs, const GtkAllocation & rhs)
{
  return (lhs.x == rhs.x) && (lhs.y == rhs.y) && 
    (lhs.width == rhs.width) && (lhs.height == rhs.height);
}

static inline bool operator!=(const GtkAllocation & lhs, const GtkAllocation & rhs)
{
  return !(lhs == rhs);
}

#ifdef GNOME_ENABLE_DEBUG
static ostream & operator<< (ostream & o, const GtkAllocation & a)
{
  o << "[x: " << a.x << " y: " << a.y << " width: " << a.width 
    << " height: " << a.height << "]";
  return o;
}
#endif

void 
PlotLayoutImpl::size_request_impl(GtkRequisition* r)
{
  g_return_if_fail(r != 0);

#ifdef GNOME_ENABLE_DEBUG
  spew();
#endif

  // Request at least 150 pixels per axis group - lame. fixme
  r->width = columns() * 200;
  r->height = rows() * 200;

  // Size request on all the contained widgets, so 
  //  we can use the request in size_allocate
  size_t i = 0;
  size_t rows = PlotLayoutImpl::rows();
  size_t columns = PlotLayoutImpl::columns();

  while ( i < rows ) {
    size_t j = 0;
    while ( j < columns ) {
      if (group_exists(i,j)) {
	Group* g = groups_[i][j];
	g_assert(g != 0);
	Gtk_Widget* p = g->get_plot();
	if ( p != 0 ) {
	  p->size_request(&REQUISITION(p));
#ifdef DISABLED_GNOME_ENABLE_DEBUG
	  g_print("Plot %u, %u requests size: %u x %u\n", 
		  i,j,REQUISITION(p).width,REQUISITION(p).height);
#endif
	}
	Axis_Base::alignment_t k = 0;
	while ( k < 4 ) {
	  Gtk_Axis* a = g->get_axis(k);
	  if ( a != 0 ) {
	    g_print("Requesting size for axis %p\n", a);
	    a->size_request(&REQUISITION(a));
#ifdef DISABLED_GNOME_ENABLE_DEBUG
	    g_print("**An axis at %u, %u requests size: %u x %u\n", 
		    i,j,REQUISITION(a).width,REQUISITION(a).height);
#endif
	  }
	  k = k + 1;
	}
      }
      ++j;
    }
    ++i;
  }
  
#ifdef GNOME_ENABLE_DEBUG
  cout << "**PlotLayoutImpl requesting size: " << r->width << " x " << r->height << endl;
#endif
}

void 
PlotLayoutImpl::size_allocate_impl(GtkAllocation* a)
{
  g_return_if_fail(a != 0);
  
#ifdef GNOME_ENABLE_DEBUG
  cout << "**" << endl;
  cout << "**PlotLayoutImpl got allocation: " << a->width << " x " << a->height << endl;
#endif
  GtkWidget* widget = GTK_WIDGET(gtkobj());

  widget->allocation = *a;
  if (GTK_WIDGET_REALIZED (widget))
    gdk_window_move_resize (widget->window,
			    a->x, 
			    a->y,
			    a->width, 
			    a->height);

#ifdef GNOME_ENABLE_DEBUG
#define PLOTPAD 10
#else
#define PLOTPAD 0
#endif

  // these are just for neatness.
  const guint16 x_origin = 0 + PLOTPAD;
  const guint16 y_origin = a->height - PLOTPAD;

  const guint16 height = a->height - PLOTPAD*2;
  const guint16 width  = a->width - PLOTPAD*2;

  const size_t rows = PlotLayoutImpl::rows();
  const size_t columns = PlotLayoutImpl::columns();

  // fixme; this shouldn't be hardcoded here. we need a way 
  //  to communicate it between the axes and here.
  static const guint16 axis_inset = 10;

#ifdef GNOME_ENABLE_DEBUG
  cout << "**Allocating size; x_origin: " << x_origin << " y_origin: " 
       << y_origin << " width: " << width << " height: " << height 
       << " rows: " << rows << " columns: " << columns 
       << endl;
#endif

  // axis dimensions
  guint16 east_axis_widths[columns];
  guint16 west_axis_widths[columns];
  guint16 north_axis_heights[rows];
  guint16 south_axis_heights[rows];
  guint16 total_vaxis_width = 0;
  guint16 total_haxis_height = 0;
  
  // get the max axis requisition for each row/column of axes.
  // We actually go over the grid twice here; it would be 
  //  nice to go only once.

  guint16 axis_size;
  size_t h = 0;
  while ( h < rows ) {
    north_axis_heights[h] = 0;
    south_axis_heights[h] = 0;

    size_t w = 0;
    while ( w < columns ) {
      Gtk_Axis* a = get_axis(h,w,Axis_Base::NORTH);
      if (a == 0) axis_size = 0; 
      else {
	axis_size = REQUISITION(a).height;
	a->set_axis_line_inset(axis_inset);
      }

      north_axis_heights[h] = MAX(north_axis_heights[h], axis_size);
	
      a = get_axis(h,w,Axis_Base::SOUTH);
      if (a == 0) axis_size = 0; 
      else {
	axis_size = REQUISITION(a).height;
	a->set_axis_line_inset(axis_inset);
      }

      south_axis_heights[h] = MAX(south_axis_heights[h], axis_size);
      ++w;
    }

    total_haxis_height += north_axis_heights[h];
    total_haxis_height += south_axis_heights[h];
    ++h;
  }

  size_t w = 0;
  while ( w < columns ) {

    east_axis_widths[w] = 0;
    west_axis_widths[w] = 0;

    size_t h = 0;
    while ( h < rows ) {
      Gtk_Axis* a = get_axis(h,w,Axis_Base::EAST);
      if (a == 0) axis_size = 0; 
      else {
	axis_size = REQUISITION(a).width;
	a->set_axis_line_inset(axis_inset);
      }

      east_axis_widths[w] = MAX(east_axis_widths[w], axis_size);
	
      a = get_axis(h,w,Axis_Base::WEST);
      if (a == 0) axis_size = 0; 
      else {
	axis_size = REQUISITION(a).width;
	a->set_axis_line_inset(axis_inset);
      }

      west_axis_widths[w] = MAX(west_axis_widths[w], axis_size);
      ++h;
    }

    total_vaxis_width += west_axis_widths[w];
    total_vaxis_width += east_axis_widths[w];
    ++w;
  }

  guint16 total_hpadding = columns*axis_inset*2;

  guint16 total_plot_width;

  if ((total_hpadding + total_vaxis_width) > width) {
    total_plot_width = 0;
  }
  else {
    total_plot_width = width - total_vaxis_width - total_hpadding;
  }

  guint16 total_vpadding = rows*axis_inset*2;
  
  guint16 total_plot_height;
  if ((total_vpadding + total_haxis_height) > height) {
    total_plot_height = 0;
  }
  else {
    total_plot_height = height - total_haxis_height - total_vpadding;
  }

  const guint16 per_plot_width   = total_plot_width / columns;
  const guint16 per_plot_height  = total_plot_height / rows;

  const guint16 per_haxis_width  = per_plot_width + axis_inset*2;
  const guint16 per_vaxis_height = per_plot_height + axis_inset*2;

  // The same remainder should apply to both plots and axes, I think.
  // (I hope, if not the tenuous attempt to keep plots and axes lined 
  //  up properly could go to hell)
  // It only applies to a single dimension of an axis though, the one 
  //  not in the above arrays.
  guint16 width_remainder  = total_plot_width % columns;
  guint16 height_remainder = total_plot_height % rows;

#ifdef GNOME_ENABLE_DEBUG
  cout << "total_vaxis_width " << total_vaxis_width << endl
       << "total_haxis_height " << total_haxis_height << endl
       << "per_haxis_width " << per_haxis_width << endl
       << "per_vaxis_height " << per_vaxis_height << endl
       << "total_plot_width "  << total_plot_width << endl
       << "total_plot_height " << total_plot_height << endl
       << "per_plot_width " << per_plot_width << endl
       << "per_plot_height " << per_plot_height << endl;
#endif
    
  

  // We are finally ready to do the allocations.
  // Row 0 starts at the bottom left, at y_origin and x_origin.
  // x_next counts up and y_next counts down.


  ///////////////// FIXME handle not-enough-space

  guint16 x_next = x_origin;

  w = 0;             // index for widths (column number)
  while ( w < columns ) {

    guint16 add_width = 0; 
    
    if (width_remainder != 0) {
      add_width = 1;
      --width_remainder;
    }
    
    guint16 this_plot_width  = per_plot_width + add_width;
    guint16 this_haxis_width  = per_haxis_width + add_width;

    guint16 y_next = y_origin;
    guint16 this_height_remainder = height_remainder;

    h = 0;     // h is the index for heights (row number)

    while ( h < rows ) {
      
      guint16 add_height = 0;
      if (this_height_remainder != 0) {
	add_height = 1;
	--this_height_remainder;
      }
      
      guint16 this_plot_height = per_plot_height + add_height;
      
      guint16 this_vaxis_height = per_vaxis_height + add_height;
      
      Gtk_Widget* plot;
      Gtk_Axis  * axis;
      GtkAllocation allocation;
      Group * group = groups_[h][w];
      
#ifdef GNOME_ENABLE_DEBUG
      cout << "**Allocating for " << h << "," << w << endl;
#endif
      
      if (group != 0) {
	plot = group->get_plot();
	
	if (plot != 0) {
	  allocation.width = this_plot_width;
	  allocation.height = this_plot_height;
	  allocation.x = x_next + west_axis_widths[w] + axis_inset;
	  allocation.y = y_next - south_axis_heights[h] - axis_inset - allocation.height;
#ifdef GNOME_ENABLE_DEBUG
	  cout << "  Plot: " << allocation << endl;
#endif
	  
	  if (allocation != ALLOCATION(plot)) {
	    plot->size_allocate(&allocation);
	  }
	}
	
	// SOUTH
	axis = group->get_axis(Axis_Base::SOUTH );
	
	if (axis != 0) {
	  allocation.width = this_haxis_width;
	  allocation.height = south_axis_heights[h];
	  allocation.x = x_next + west_axis_widths[w];
	  allocation.y = y_next - allocation.height;
#ifdef GNOME_ENABLE_DEBUG
	  cout << "  South: " << allocation << " realized: " << axis->realized() << endl; 
#endif
	  if (allocation != ALLOCATION(axis)) {
	    axis->size_allocate(&allocation);
	  }
	}
	// SOUTH END
			 
	// WEST
	axis = group->get_axis(Axis_Base::WEST );
	
	if (axis != 0) {
	  allocation.width = west_axis_widths[w];
	  allocation.height = this_vaxis_height;
	  allocation.x = x_next;
	  allocation.y = y_next - south_axis_heights[h] - 
	    allocation.height;
#ifdef GNOME_ENABLE_DEBUG
	  cout << "  West: " << allocation  << " realized: " << axis->realized() << endl; 
#endif
	  if (allocation != ALLOCATION(axis)) {
	    axis->size_allocate(&allocation);
	  }
	}
	// WEST END

	// NORTH
	axis = group->get_axis(Axis_Base::NORTH );
	
	if (axis != 0) {
	  allocation.width = this_haxis_width;
	  allocation.height = north_axis_heights[h];
	  allocation.x = x_next + west_axis_widths[w];
	  allocation.y = y_next - south_axis_heights[h] - 
	    this_vaxis_height - allocation.height;
#ifdef GNOME_ENABLE_DEBUG
	  cout << "  North: " << allocation <<  " realized: " << axis->realized() << endl; 
#endif
	  if (allocation != ALLOCATION(axis)) {
	    axis->size_allocate(&allocation);
	  }
	}
	// NORTH END
	
	// EAST
	axis = group->get_axis(Axis_Base::EAST );
	
	if (axis != 0) {
	  allocation.width = east_axis_widths[w];
	  allocation.height = this_vaxis_height;
	  allocation.x = x_next + this_haxis_width + west_axis_widths[w];
	  allocation.y = y_next - south_axis_heights[h] - allocation.height;
#ifdef GNOME_ENABLE_DEBUG
	  cout << "  East: " << allocation << " realized: " << axis->realized() << endl; 
#endif
	  if (allocation != ALLOCATION(axis)) {
	    axis->size_allocate(&allocation);
	  }
	}
	// EAST_END
      }

      y_next -= (south_axis_heights[h] + this_vaxis_height + north_axis_heights[h]);

      ++h;
    }

    x_next += (east_axis_widths[w] + this_haxis_width + west_axis_widths[w]); 
    ++w;
  }
#ifdef GNOME_ENABLE_DEBUG
  cout << "**" << endl;
#endif
}

void
PlotLayoutImpl::realize_impl()
{
  GdkWindowAttr attributes;
  gint attributes_mask;
  GtkWidget* widget = GTK_WIDGET(gtkobj());

  GTK_WIDGET_SET_FLAGS (widget, GTK_REALIZED);

  attributes.window_type = GDK_WINDOW_CHILD;
  attributes.x = widget->allocation.x;
  attributes.y = widget->allocation.y;
  attributes.width = widget->allocation.width;
  attributes.height = widget->allocation.height;
  attributes.wclass = GDK_INPUT_OUTPUT;
  attributes.visual = gtk_widget_get_visual (widget);
  attributes.colormap = gtk_widget_get_colormap (widget);
  attributes.event_mask = gtk_widget_get_events (widget);
  attributes.event_mask |= GDK_EXPOSURE_MASK;

  attributes_mask = GDK_WA_X | GDK_WA_Y | GDK_WA_VISUAL | GDK_WA_COLORMAP;

  widget->window = gdk_window_new (gtk_widget_get_parent_window (widget), &attributes, 
				   attributes_mask);
  gdk_window_set_user_data (widget->window, widget);

  widget->style = gtk_style_attach (widget->style, widget->window);
  gtk_style_set_background (widget->style, widget->window, GTK_STATE_NORMAL);

  // Realize children

  size_t i = 0;
  size_t r = rows();
  size_t c = columns();
#ifdef GNOME_ENABLE_DEBUG
  cout << "Realizing with " << r << " rows and " << c << " columns." << endl;
#endif
  while ( i < r ) {
    size_t j = 0;
    while ( j < c ) {
      Group* g = groups_[i][j];
      if (g != 0) {
	Gtk_Widget* realize_us[5];
	realize_us[0] = g->get_plot();
	realize_us[1] = g->get_axis(Axis_Base::NORTH);
	realize_us[2] = g->get_axis(Axis_Base::SOUTH);
	realize_us[3] = g->get_axis(Axis_Base::EAST);
	realize_us[4] = g->get_axis(Axis_Base::WEST);

	size_t k = 0;
	while ( k < 5 ) {
	  if (realize_us[k] != 0) {
	    if (realize_us[k]->visible() && ! realize_us[k]->realized()) {
	      realize_us[k]->realize();
	      g_print("Realized child: %p\n", realize_us[k]);
	    }
	  }
	  ++k;
	}
      }
      ++j;
    }			
    ++i;    
  }
}

void
PlotLayoutImpl::map_impl() 
{
  GtkWidget* widget = GTK_WIDGET(gtkobj());

  GTK_WIDGET_SET_FLAGS (widget, GTK_MAPPED);

  gdk_window_show (widget->window);
	
  size_t i = 0;
  size_t r = rows();
  size_t c = columns();
#ifdef GNOME_ENABLE_DEBUG
  cout << "Mapping with " << r << " rows and " << c << " columns." << endl;
#endif
  while ( i < r ) {
    size_t j = 0;
    while ( j < c ) {
      Group* g = groups_[i][j];
      if (g != 0) {
	Gtk_Widget* map_us[5];
	map_us[0] = g->get_plot();
	map_us[1] = g->get_axis(Axis_Base::NORTH);
	map_us[2] = g->get_axis(Axis_Base::SOUTH);
	map_us[3] = g->get_axis(Axis_Base::EAST);
	map_us[4] = g->get_axis(Axis_Base::WEST);

	size_t k = 0;
	while ( k < 5 ) {
	  if (map_us[k] != 0) {
	    if (map_us[k]->visible() && ! map_us[k]->mapped()) {
	      map_us[k]->map();
	      g_print("Mapped child: %p\n", map_us[k]);
	    }
	  }
	  ++k;
	}
      }
      ++j;
    }			
    ++i;    
  }
}

void
PlotLayoutImpl::unmap_impl()
{
  g_print("Unmapping!\n");
  GtkWidget* widget = GTK_WIDGET(gtkobj());

  GTK_WIDGET_UNSET_FLAGS (widget, GTK_MAPPED);
}

void
PlotLayoutImpl::paint(GdkRectangle* area)
{
  GtkWidget* widget = GTK_WIDGET(gtkobj());
  
  size_t r = rows();
  size_t c = columns();
  size_t i = 0;
  while ( i < r ) {
    size_t j = 0;
    while ( j < c ) {
      Gtk_Widget* plot = get_plot(i,j);
      if (plot != 0) {
	// We want to draw a black square around the plots.
	// Clearly it would be more correct to draw this only
	//  when necessary instead of ignoring 'area'.
	gdk_draw_rectangle (widget->window,
			    widget->style->black_gc, 
			    FALSE,
			    ALLOCATION(plot).x-1, ALLOCATION(plot).y-1, 
			    ALLOCATION(plot).width+1, ALLOCATION(plot).height+1);
      }
#ifdef DISABLED_GNOME_ENABLE_DEBUG
      Gtk_Widget* axis;
      axis = get_axis(i,j,Axis_Base::NORTH);
      if (axis != 0) {
	gdk_draw_rectangle (widget->window,
			    widget->style->black_gc, 
			    FALSE,
			    ALLOCATION(axis).x-1, ALLOCATION(axis).y-1, 
			    ALLOCATION(axis).width+1, ALLOCATION(axis).height+1);
      }
      axis = get_axis(i,j,Axis_Base::SOUTH);
      if (axis != 0) {
	gdk_draw_rectangle (widget->window,
			    widget->style->black_gc, 
			    FALSE,
			    ALLOCATION(axis).x-1, ALLOCATION(axis).y-1, 
			    ALLOCATION(axis).width+1, ALLOCATION(axis).height+1);
      }
      axis = get_axis(i,j,Axis_Base::EAST);
      if (axis != 0) {
	gdk_draw_rectangle (widget->window,
			    widget->style->black_gc, 
			    FALSE,
			    ALLOCATION(axis).x-1, ALLOCATION(axis).y-1, 
			    ALLOCATION(axis).width+1, ALLOCATION(axis).height+1);
      }
      axis = get_axis(i,j,Axis_Base::WEST);
      if (axis != 0) {
	gdk_draw_rectangle (widget->window,
			    widget->style->black_gc, 
			    FALSE,
			    ALLOCATION(axis).x-1, ALLOCATION(axis).y-1, 
			    ALLOCATION(axis).width+1, ALLOCATION(axis).height+1);
      }
#endif

      ++j;
    }
    ++i;
  }
}

gint
PlotLayoutImpl::expose_event_impl(GdkEventExpose* e)
{
  if (drawable()) {
    paint(&e->area);
  }
  // We don't support windowless children.
  return FALSE;
}

void
PlotLayoutImpl::draw_impl(GdkRectangle* area)
{
  g_print("Drawing! x: %u y: %u w: %u h: %u\n", 
	  area->x, area->y, area->width, area->height); 

  if (drawable()) {
    gdk_window_clear_area (get_window(),
			   area->x, area->y,
			   area->width, area->height);
    paint(area);
  }

  size_t i = 0;
  size_t r = rows();
  size_t c = columns();
  while ( i < r ) {
    size_t j = 0;
    while ( j < c ) {
      Group* g = groups_[i][j];
      if (g != 0) {
	Gtk_Widget* draw_us[5];
        draw_us[0] = g->get_plot();
	draw_us[1] = g->get_axis(Axis_Base::NORTH);
	draw_us[2] = g->get_axis(Axis_Base::SOUTH);
	draw_us[3] = g->get_axis(Axis_Base::EAST);
	draw_us[4] = g->get_axis(Axis_Base::WEST);

	size_t k = 0;
	while ( k < 5 ) {
	  if (draw_us[k] != 0) {
	    GdkRectangle child_area;

	    if (draw_us[k]->intersect(area, &child_area)) {
	      draw_us[k]->draw(&child_area);
	      g_print("Drew child: %p\n", draw_us[k]);
	    }
	  }
	  ++k;
	}
      }
      ++j;
    }			
    ++i;    
  }
}

gint
PlotLayoutImpl::configure_event_impl(GdkEventConfigure* e)
{
  queue_resize();
  return FALSE;
}

void 
PlotLayoutImpl::setup_child(Gtk_Widget* child)
{
  g_return_if_fail(child != 0);
  child->set_parent(this);
  if (realized() && !child->realized()) {
    child->realize();
    g_print("Realized child: %p\n", child);
  }
  if (mapped() && !child->mapped()) {
    child->map();
    g_print("Mapped child: %p\n", child);
  }
  if (child->visible() && visible()) {
    queue_resize();
  }
}

void
PlotLayoutImpl::clear()
{
  size_t i = 0;
  size_t r = rows();
  size_t c = columns();
  while ( i < r ) {
    size_t j = 0;
    while ( j < c ) {
      delete groups_[i][j];
      groups_[i][j] = 0;
      ++j;
    }
    ++i;
  }
  queue_resize();
}

// These calls destroy any existing plot or axis
void
PlotLayoutImpl::place_plot(Gtk_Widget* w, size_t row, size_t col)
{
#ifdef GNOME_ENABLE_DEBUG
  cout << "Placing plot " << row << "," << col << endl;
#endif

  ensure_exists(row,col);

  groups_[row][col]->replace_plot(w);

  setup_child(w);

#ifdef GNOME_ENABLE_DEBUG
  spew();
#endif
}

void
PlotLayoutImpl::place_axis(Gtk_Axis* w, size_t row, size_t col, Axis_Base::alignment_t a)
{
#ifdef GNOME_ENABLE_DEBUG
  cout << "Placing axis " << row << "," << col << endl;
#endif

  ensure_exists(row,col);

  groups_[row][col]->replace_axis(w,a);

  setup_child(w);
#ifdef GNOME_ENABLE_DEBUG
  spew();
#endif
}

// Destroy everything on a row or column and shift
//  other rows/columns accordingly.
void 
PlotLayoutImpl::remove_row(size_t row)
{
  g_return_if_fail(row < rows());

  size_t i = 0;
  while (i < columns()) {
    erase_group(row,i);
    ++i;
  }
  groups_.delete_row(row);

  queue_resize();
}

void 
PlotLayoutImpl::remove_column(size_t col)
{
  g_return_if_fail(col < columns());

  size_t i = 0;
  while (i < rows()) {
    erase_group(i,col);
    ++i;
  }
  groups_.delete_column(col);

  queue_resize();
}

  
// These don't shift other things, they just leave a blank.
void
PlotLayoutImpl::erase_axis(size_t row, size_t col, Axis_Base::alignment_t a)
{
  g_return_if_fail(in_range(row,col));

  if (groups_[row][col] == 0) return;
  groups_[row][col]->replace_axis(0,a);

  queue_resize();
}

void
PlotLayoutImpl::erase_plot(size_t row, size_t col)
{
  g_return_if_fail(in_range(row,col));

  if (groups_[row][col] == 0) return;
  groups_[row][col]->replace_plot(0);

  queue_resize();
}

void
PlotLayoutImpl::erase_axes(size_t row, size_t col)
{
  g_return_if_fail(in_range(row,col));

  if (groups_[row][col] == 0) return;
  groups_[row][col]->erase_axes();

  queue_resize();
}

// delete both axes and plots
void
PlotLayoutImpl::erase_group(size_t row, size_t col)
{
  g_return_if_fail(in_range(row,col));

  if (groups_[row][col] == 0) return;
  groups_[row][col]->clear();

  queue_resize();
}
  
  
void 
PlotLayoutImpl::move_axis (size_t row1, size_t col1, size_t row2, size_t col2, 
			   Axis_Base::alignment_t a)
{
  g_return_if_fail(in_range(row1,col1));
  
  if (row1 == col1 && row2 == col2) return;

  ensure_exists(row2,col2);
  
  erase_axis(row2,col2,a);
  
  Group* src = groups_[row1][col1];

  // we ensured dest exists.
  Group* dest = groups_[row2][col2]; 
  if (src != 0) {
    dest->set_axis(src->get_axis(a),a);
    src->set_axis(0,a);
  }
  else {
    // shouldn't happen really
    dest->set_axis(0,a);
  }
}

void
PlotLayoutImpl::move_plot (size_t row1, size_t col1, size_t row2, size_t col2)
{
  g_return_if_fail(in_range(row1,col1));
  
  if (row1 == col1 && row2 == col2) return;

  ensure_exists(row2,col2);

  erase_plot(row2,col2);
  
  Group* src = groups_[row1][col1];

  // we ensured dest exists.
  Group* dest = groups_[row2][col2]; 
  if (src != 0) {
    dest->set_plot(src->get_plot());
    src->set_plot(0);
  }
  else {
    // shouldn't happen really
    dest->set_plot(0);
  }
}

void
PlotLayoutImpl::move_group(size_t row1, size_t col1, size_t row2, size_t col2)
{
  g_return_if_fail(in_range(row1,col1));

  if (row1 == col1 && row2 == col2) return;
  
  ensure_exists(row2,col2);

  erase_group(row2,col2);
  groups_[row2][col2] = groups_[row1][col1];
  groups_[row1][col1] = 0;

  queue_resize();
}

void
PlotLayoutImpl::swap_group(size_t row1, size_t col1, size_t row2, size_t col2)
{
  g_return_if_fail(in_range(row1,col1));
  g_return_if_fail(in_range(row2,col2));

  if (row1 == col1 && row2 == col2) return;

  Group* tmp = groups_[row1][col1];
  groups_[row1][col1] = groups_[row2][col2];
  groups_[row2][col2] = tmp;

  queue_resize();
}

// Check on state
bool 
PlotLayoutImpl::plot_exists(size_t row, size_t col)
{
  if (!group_exists(row,col)) return false;
  if (groups_[row][col]->get_plot() == 0) return false;
  else return true;
}

bool 
PlotLayoutImpl::axis_exists(size_t row, size_t col, Axis_Base::alignment_t a)
{
  if (!group_exists(row,col)) return false;
  if (groups_[row][col]->get_axis(a) == 0) return false;
  else return true;
}

bool 
PlotLayoutImpl::group_exists(size_t row, size_t col)
{
  if (!in_range(row,col)) return false;
  if (groups_[row][col] == 0) return false;
  else return true;
}


// per-plot title
void
PlotLayoutImpl::set_title(size_t row, size_t col, const string & title)
{

  queue_resize();
}

// whole-grid title.
void
PlotLayoutImpl::set_title(const string & title)
{

  queue_resize();
}


Gtk_Widget* 
PlotLayoutImpl::get_plot(size_t row, size_t col)
{
  g_return_val_if_fail(in_range(row,col),0);
  if (groups_[row][col] != 0) return groups_[row][col]->get_plot();
  else return 0;
}

Gtk_Axis*   
PlotLayoutImpl::get_axis(size_t row, size_t col, Axis_Base::alignment_t a)
{
  g_return_val_if_fail(in_range(row,col),0);
  if (groups_[row][col] != 0) return groups_[row][col]->get_axis(a);
  else return 0;
}


size_t 
PlotLayoutImpl::rows() const
{
  return groups_.rows();
}
 
size_t 
PlotLayoutImpl::columns() const
{
  return groups_.columns();
}
 



inline void PlotLayout::clear() {
  impl_->clear(); 
} 


inline void PlotLayout::place_plot(Gtk_Widget* w, size_t row, size_t col) {
  impl_->place_plot(w,row,col);
}
inline void PlotLayout::place_axis(Gtk_Axis* w, size_t row, size_t col, Axis_Base::alignment_t a) {
  impl_->place_axis(w,row,col,a);
}

inline void PlotLayout::remove_row(size_t row) {
  impl_->remove_row(row);
}
inline void PlotLayout::remove_column(size_t col) {
  impl_->remove_column(col);
}

inline void PlotLayout::erase_axis(size_t row, size_t col, Axis_Base::alignment_t a) {
  impl_->erase_axis(row,col,a);
}
inline void PlotLayout::erase_plot(size_t row, size_t col) {
  impl_->erase_plot(row,col);
}
inline void PlotLayout::erase_axes(size_t row, size_t col) {
  impl_->erase_axes(row,col);
}

inline void PlotLayout::erase_group(size_t row, size_t col) {
  impl_->erase_group(row,col);
}  

inline void PlotLayout::move_axis (size_t row1, size_t col1, size_t row2, size_t col2, 
				   Axis_Base::alignment_t a)
{
  impl_->move_axis(row1,col1,row2,col2,a);
}

inline void PlotLayout::move_plot (size_t row1, size_t col1, size_t row2, size_t col2)
{
  impl_->move_plot(row1,col1,row2,col2);
}

inline void PlotLayout::move_group(size_t row1, size_t col1, size_t row2, size_t col2) {
  impl_->move_group(row1,col1,row2,col2);
}
inline void PlotLayout::swap_group(size_t row1, size_t col1, size_t row2, size_t col2) {
  impl_->swap_group(row1,col1,row2,col2);
}

inline bool PlotLayout::plot_exists(size_t row, size_t col) {
  return impl_->plot_exists(row,col);
}
inline bool PlotLayout::axis_exists(size_t row, size_t col, Axis_Base::alignment_t a) {
  return impl_->axis_exists(row,col,a);
}
inline bool PlotLayout::group_exists(size_t row, size_t col) {
  return impl_->group_exists(row,col);
}


inline void PlotLayout::set_title(size_t row, size_t col, const string & title) {
  impl_->set_title(row,col,title);
}

inline void PlotLayout::set_title(const string & title) {
  impl_->set_title(title);
}

inline Gtk_Widget* PlotLayout::get_plot(size_t row, size_t col) {
  return impl_->get_plot(row,col);
}

inline Gtk_Axis*   PlotLayout::get_axis(size_t row, size_t col, Axis_Base::alignment_t a) {
  return impl_->get_axis(row,col,a);
}

inline size_t PlotLayout::rows() const {
  return impl_->rows();
} 

inline size_t PlotLayout::columns() const {
  return impl_->columns();
} 





