// -*- C++ -*-

/* 
 * rubberband.cc
 *
 * Copyright (C) 1998 EMC Capital Management, Inc.
 *
 * Developed by Jon Trowbridge <trow@emccta.com> and
 * Havoc Pennington <hp@emccta.com>.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

//////////////////////////////////////////////////////////////
///////////// WARNING nearly everything in this file 
/////////////  is very sensitive to order. Don't reorder
/////////////  the lines unless you know you want to.

#include "rubberband.h"

#include <math.h>

RubberBand::RubberBand() : x_origin_(0), y_origin_(0),
  x_(0), y_(0), motion_cnxn_(), release_cnxn_(), 
  time_(0), stretching_(false), draw_(0), 
  width_locked_(false), height_locked_(false), 
  width_lock_(-1), height_lock_(-1)
{

}

RubberBand::~RubberBand()
{
  if (stretching_) stop();
}

void
RubberBand::start(Gdk_Draw* d, Gtk_Widget* w, GdkEventButton* be)
{
  g_return_if_fail(d != 0);
  g_return_if_fail(w != 0);

  if (stretching_) return;

  x_ = x_origin_ = (gint) be->x;
  y_ = y_origin_ = (gint) be->y;
  draw_ = d;

  gdk_pointer_grab(w->gtkobj()->window, FALSE, 
		   (GdkEventMask) 
		   (GDK_POINTER_MOTION_HINT_MASK | 
		    GDK_BUTTON1_MOTION_MASK | 
		    GDK_BUTTON_RELEASE_MASK | 
		    GDK_BUTTON_PRESS_MASK),
		   w->gtkobj()->window, // confine to this window
		   NULL,   // cursor
		   be->time);

  stretching_ = true;
  time_ = be->time;
  
  draw_->gc_save();
  draw_->gc()->set_function(GDK_XOR);
  draw_->gc()->set_foreground("white");
  draw();

  release_cnxn_ = 
    connect_to_method(w->button_release_event, this, 
		      &RubberBand::button_release_cb);
  
  motion_cnxn_ = 
    connect_to_method(w->motion_notify_event, this, 
		      &RubberBand::motion_notify_cb);

  press_cnxn_ = 
    connect_to_method(w->button_press_event, this,
		      &RubberBand::button_press_cb);
  
  g_print("start\n");
}

void
RubberBand::stop()
{
  if (!stretching_) return;
  
  // We want to stop immediately,
  //  without processing future events.
  motion_cnxn_.disconnect();
  release_cnxn_.disconnect();
  press_cnxn_.disconnect();

  // Clean up any events, release the pointer.
  gdk_pointer_ungrab(time_);
  gdk_flush();

  undraw();
  
  draw_->gc_restore();
  draw_ = 0;

  stretching_ = false;

  g_print("stop\n");
}

void 
RubberBand::lock_width(gint w)
{
  width_locked_ = true;
  width_lock_ = w;
}

void
RubberBand::lock_height(gint h)
{
  height_locked_ = true;
  height_lock_ = h;
}

void 
RubberBand::unlock()
{
  width_locked_ = false;
  height_locked_ = false;
}

void
RubberBand::draw()
{
  g_return_if_fail(stretching_);
  g_return_if_fail(draw_ != 0);

  // Allow locking to be changed based
  //  on current position.
  changed(x_origin_,y_origin_,x_,y_);

  gint x;
  if (width_locked_) x = MIN(x_origin_ + width_lock_, x_origin_);
  else x = MIN(x_origin_,x_);
  gint y;
  if (height_locked_) y = MIN(y_origin_ + height_lock_, y_origin_);
  else y = MIN(y_origin_,y_);

  // origin + negative lock can be less than 0
  if (x < 0) x = 0;
  if (y < 0) y = 0;

  gint width;
  if (width_locked_) width = abs(width_lock_);
  else width = abs(x_origin_ - x_);

  gint height;

  if (height_locked_) height = abs(height_lock_);
  else height = abs(y_origin_ - y_);

  draw_->draw_rectangle(false, x, y, width, height);

  //  g_print("RB Draw: %d, %d  %d x %d\n", x, y, width, height);

  last_.x = x;
  last_.y = y;
  last_.width = width;
  last_.height =  height;
}

void
RubberBand::undraw()
{
  g_return_if_fail(stretching_);
  g_return_if_fail(draw_ != 0);

  draw_->draw_rectangle(false, last_.x, last_.y, last_.width, last_.height);
  //  g_print(" Undraw: %d, %d  %d x %d\n", last_.x, last_.y, last_.width, last_.height);
}

gint
RubberBand::motion_notify_cb(GdkEventMotion* me)
{
  g_return_val_if_fail(stretching_, FALSE);

  GdkModifierType state;

  if (me->is_hint)
    gdk_window_get_pointer (me->window, &x_, &y_, &state);
  else
    {
      x_ = (gint) me->x;
      y_ = (gint) me->y;
      state = (GdkModifierType) me->state;
    }

  // I guess we don't actually care about the state fixme

  undraw();
  draw();

  return TRUE;
}

gint
RubberBand::button_release_cb(GdkEventButton* be)
{
  time_ = be->time;
  stop();
  // have to snap after GC is reset.
  snapped(last_.x, last_.y,
	  last_.width, last_.height);
  return TRUE;
}

gint 
RubberBand::button_press_cb(GdkEventButton* be)
{
  // Any button besides button 1 (The drag button) 
  //  cancels the operation.
  if (be->button != 1) {
    stop();
  }  
  return TRUE;
}

